#include "darwin_leg_kinematics.h"
#include "darwin_robot_exceptions.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <iostream>
#ifdef _HAVE_XSD
#include "xml/darwin_kin_chain.hxx"
#endif

CDarwinLegKinematics::CDarwinLegKinematics()
{
  srand(time(NULL));  
  this->num_dof=0;
  this->fwd_kin_transforms.clear();
  this->fwd_kin_params.clear();
}

double CDarwinLegKinematics::angle_norm(double angle)
{
  while(fabs(angle)>1.5707)
  {
    if(angle>0) angle-=1.5707;
    if(angle<0) angle+=1.5707;
  }

  return angle;
}

Eigen::Matrix4d CDarwinLegKinematics::inv_kin_gen_transform(Eigen::Vector3d &pos,Eigen::Vector3d &rot)
{
  Eigen::Matrix4d transform=Eigen::Matrix4d::Identity();
  double Cx = cos(rot(0));
  double Cy = cos(rot(1));
  double Cz = cos(rot(2));
  double Sx = sin(rot(0));
  double Sy = sin(rot(1));
  double Sz = sin(rot(2));

  transform(0,0)=Cz*Cy;
  transform(0,1)=Cz*Sy*Sx-Sz*Cx;
  transform(0,2)=Cz*Sy*Cx+Sz*Sx;
  transform(0,3)=pos(0);
  transform(1,0)=Sz*Cy;
  transform(1,1)=Sz*Sy*Sx+Cz*Cx;
  transform(1,2)=Sz*Sy*Cx-Cz*Sx;
  transform(1,3)=pos(1);
  transform(2,0)=-Sy;
  transform(2,1)=Cy*Sx;
  transform(2,2)=Cy*Cx;
  transform(2,3)=pos(2);

  return transform;
}

void CDarwinLegKinematics::zero_small_values(Eigen::Matrix4d &matrix,double threshold)
{
  unsigned int i,j;

  for(i=0;i<4;i++) 
    for(j=0;j<4;j++)
      if(fabs(matrix(i,j))<threshold)
        matrix(i,j)=0.0;
}

void CDarwinLegKinematics::load_chain(std::string &filename)
{
  joints_t::joint_iterator iterator;
  Eigen::Matrix4d transform;
  struct stat buffer;
  TDHParams params;
  unsigned int i=0;

  if(stat(filename.c_str(),&buffer)==0)
  {
    try{
      std::auto_ptr<joints_t> chain(darwin_kin_chain(filename.c_str(), xml_schema::flags::dont_validate));
      this->fwd_kin_params.clear();
      this->fwd_kin_transforms.clear();
      this->num_dof=0;
      for(iterator=chain->joint().begin(),i=0;iterator!=chain->joint().end();iterator++,i++)
      {
        params.alpha=iterator->params().alpha();
        params.a=iterator->params().a();
        params.d=iterator->params().d();
        params.theta=iterator->params().theta();
        this->fwd_kin_params.push_back(params);
        // precompute the transforms
        transform(2,0)=0.0;
        transform(2,1)=sin(params.alpha);
        transform(2,2)=cos(params.alpha);
        transform(2,3)=params.d;
        transform(3,0)=0.0;
        transform(3,1)=0.0;
        transform(3,2)=0.0;
        transform(3,3)=1.0;
        this->fwd_kin_transforms.push_back(transform);
        this->num_dof++;
      }
    }catch(const xml_schema::exception& e){
      std::ostringstream os;
      os << e;
      throw CDarwinRobotException(_HERE_,os.str());
    }
  }
  else
    throw CDarwinRobotException(_HERE_,"Leg chain definition file does not exist");
}

void CDarwinLegKinematics::get_forward_kinematics(std::vector<double> &angles,std::vector<double> &pos,std::vector<double> &rot)
{
  unsigned int i=0;
  Eigen::Matrix4d total_transform=Eigen::Matrix4d::Identity(),rotate;

  if(angles.size()!=this->num_dof)
    throw CDarwinRobotException(_HERE_,"Invalid angle vector size");
  else
  {
    for(i=0;i<this->num_dof;i++)
    {
      this->fwd_kin_transforms[i](0,0)=cos(angles[i]+this->fwd_kin_params[i].theta);
      this->fwd_kin_transforms[i](0,1)=-sin(angles[i]+this->fwd_kin_params[i].theta)*cos(this->fwd_kin_params[i].alpha);
      this->fwd_kin_transforms[i](0,2)=sin(angles[i]+this->fwd_kin_params[i].theta)*sin(this->fwd_kin_params[i].alpha);
      this->fwd_kin_transforms[i](0,3)=this->fwd_kin_params[i].a*cos(angles[i]+this->fwd_kin_params[i].theta);
      this->fwd_kin_transforms[i](1,0)=sin(angles[i]+this->fwd_kin_params[i].theta);
      this->fwd_kin_transforms[i](1,1)=cos(angles[i]+this->fwd_kin_params[i].theta)*cos(this->fwd_kin_params[i].alpha);
      this->fwd_kin_transforms[i](1,2)=-cos(angles[i]+this->fwd_kin_params[i].theta)*sin(this->fwd_kin_params[i].alpha);
      this->fwd_kin_transforms[i](1,3)=this->fwd_kin_params[i].a*sin(angles[i]+this->fwd_kin_params[i].theta);
      total_transform*=this->fwd_kin_transforms[i];
    }
    rotate << 0.0,0.0,1.0,0.0,
              0.0,1.0,0.0,0.0,
              -1.0,0.0,0.0,0.0,
              0.0,0.0,0.0,1.0;
    total_transform*=rotate.inverse();
    this->zero_small_values(total_transform,0.001);

    pos.resize(3);
    pos[0]=total_transform(0,3);
    pos[1]=total_transform(1,3);
    pos[2]=total_transform(2,3);
    rot.resize(3);
    if(fabs(total_transform(0,0))<0.001 && fabs(total_transform(1,0))<0.001)
    {
      rot[0]=atan2(total_transform(0,1),total_transform(1,1));
      rot[1]=asin(-total_transform(2,0));
      rot[2]=0.0;
    }
    else
    {
      rot[0]=atan2(total_transform(2,1),total_transform(2,2));
      rot[1]=atan2(-total_transform(2,0),sqrt(pow(total_transform(0,0),2)+pow(total_transform(1,0),2)));  
      rot[2]=atan2(total_transform(1,0),total_transform(0,0));
    }
  }
}

void CDarwinLegKinematics::get_inverse_kinematics(std::vector<double> &pos,std::vector<double> &rot,std::vector<double> &angles)
{
  Eigen::Matrix4d Tad,Tda,Tcd,Tdc,Tac;
  Eigen::Vector3d vec,orientation,position;
  double _Rac,_Acos,_Atan,_k,_l,_m,_n,_s,_c,_theta,tmp;

  angles.resize(6);

  orientation << rot[0], rot[1], rot[2];
  position << pos[0], pos[1], pos[2]+DARWIN_PELVIS_LENGTH;
  Tad=this->inv_kin_gen_transform(position,orientation);

  vec(0)=pos[0]+Tad(0,2)*DARWIN_ANKLE_LENGTH;
  vec(1)=pos[1]+Tad(1,2)*DARWIN_ANKLE_LENGTH;
  vec(2)=(pos[2]+DARWIN_PELVIS_LENGTH)+Tad(2,2)*DARWIN_ANKLE_LENGTH;

  // Get Knee
  _Rac=vec.norm();

  tmp=(pow(_Rac,2.0)-pow(DARWIN_THIGH_LENGTH,2.0)-pow(DARWIN_CALF_LENGTH,2.0))/(2.0*DARWIN_THIGH_LENGTH*DARWIN_CALF_LENGTH);
  if(tmp>1.0) tmp=1.0;
  else if(tmp<-1.0) tmp=1.0;
  _Acos=acos(tmp);
  if(isnan(_Acos)==1)
    throw CDarwinRobotException(_HERE_,"Impossible to find solution"); 
  // this is not the same for both legs
  angles[3]=_Acos;

  // Get Ankle Roll
  if(Tad.determinant()==0.0)
    throw CDarwinRobotException(_HERE_,"Impossible to find solution"); 
  Tda=Tad.inverse();
  _k=sqrt(pow(Tda(1,3),2.0)+pow(Tda(2,3),2.0));
  _l=sqrt(pow(Tda(1,3),2.0)+pow(Tda(2,3)-DARWIN_ANKLE_LENGTH,2.0));
  _m=(pow(_k,2.0)-pow(_l,2.0)-pow(DARWIN_ANKLE_LENGTH,2.0))/(2.0*_l*DARWIN_ANKLE_LENGTH);
  if(_m>1.0) _m=1.0;
  else if(_m<-1.0) _m=-1.0;
  _Acos=acos(_m);
  if(isnan(_Acos)==1)
    throw CDarwinRobotException(_HERE_,"Impossible to find solution"); 
  if(Tda(1,3)<0.0)
    angles[5]=-_Acos;
  else
    angles[5]=_Acos;

  // Get Hip Yaw
  orientation << angles[5],0,0;
  position << 0,0,-DARWIN_ANKLE_LENGTH;
  Tcd=this->inv_kin_gen_transform(position,orientation);
  if(Tcd.determinant()==0.0)
    throw CDarwinRobotException(_HERE_,"Impossible to find solution"); 
  Tdc=Tcd.inverse();
  Tac=Tad*Tdc;

  _Atan=atan2(-Tac(0,1),Tac(1,1));
  if(isinf(_Atan)==1)
    throw CDarwinRobotException(_HERE_,"Impossible to find solution"); 
  angles[0]=_Atan;

  // Get Hip Roll
  _Atan=atan2(Tac(2,1),-Tac(0,1)*sin(angles[0])+Tac(1,1)*cos(angles[0]));
  if(isinf(_Atan)==1)
    throw CDarwinRobotException(_HERE_,"Impossible to find solution"); 
  angles[1]=_Atan;

  // Get Hip Pitch and Ankle Pitch
  _Atan=atan2(Tac(0,2)*cos(angles[0])+Tac(1,2)*sin(angles[0]),Tac(0,0)*cos(angles[0])+Tac(1,0)*sin(angles[0]));
  if(isinf(_Atan)==1)
    throw CDarwinRobotException(_HERE_,"Impossible to find solution"); 
  _theta=_Atan;
  _k=sin(angles[3])*DARWIN_CALF_LENGTH;
  _l=-DARWIN_THIGH_LENGTH-cos(angles[3])*DARWIN_CALF_LENGTH;
  _m=cos(angles[0])*vec(0)+sin(angles[0])*vec(1);
  _n=cos(angles[1])*vec(2)+sin(angles[0])*sin(angles[1])*vec(0)-cos(angles[0])*sin(angles[1])*vec(1);
  _s=(_k*_n+_l*_m)/(pow(_k,2.0)+pow(_l,2.0));
  _c=(_n-_k*_s)/_l;
  _Atan=atan2(_s,_c);
  if(isinf(_Atan)== 1)
    throw CDarwinRobotException(_HERE_,"Impossible to find solution"); 
  angles[2]=_Atan;
  angles[4]=-(_theta-angles[3]-angles[2]);
}

CDarwinLegKinematics::~CDarwinLegKinematics()
{
  this->num_dof=0;
  this->fwd_kin_transforms.clear();
  this->fwd_kin_params.clear();
}

