#ifndef _DARWIN_ROBOT_H
#define _DARWIN_ROBOT_H

#include "dynamixel.h"
#include "dynamixelserver_ftdi.h"
#include "darwin_registers.h"

#define           MAX_NUM_SERVOS        31

extern const std::string servo_names[MAX_NUM_SERVOS];
/* available motion modules */
typedef enum {DARWIN_MM_NONE=0x00,DARWIN_MM_ACTION=0x01,DARWIN_MM_WALKING=0x02,DARWIN_MM_JOINTS=0x03,DARWIN_MM_HEAD=0x04,DARWIN_MM_GRIPPER=0x05} mm_mode_t;
/* available grippers */
typedef enum {LEFT_GRIPPER=0,RIGHT_GRIPPER=1} grippers_t;
/* available leds */
#define           GPIO_NUM_LEDS         11

typedef enum {LED_TX=0,LED_RX=1,LED_2=2,LED_3=3,LED_4=4,LED_5_R=5,LED_5_G=6,LED_5_B=7,LED_6_R=8,LED_6_G=9,LED_6_B=10} led_t;

typedef struct
{
  unsigned short int control_reg;
  unsigned short int data_reg;
  unsigned char value_mask;
  unsigned char toggle_mask;
  unsigned char blink_mask;
}led_info_t;

/* available pushbuttons */
#define           GPIO_NUM_PB           2

typedef enum {START_PB=0,MODE_PB=1} pushbutton_t;

#define           ADC_NUM_CHANNELS      16

typedef enum {ADC_CH1=0,ADC_CH2=1,ADC_CH3=2,ADC_CH4=3,ADC_CH5=4,ADC_CH6=5,ADC_CH7=6,ADC_CH8=7,
              ADC_CH9=8,ADC_CH10=9,ADC_CH12=10,ADC_CH14=11} adc_t;

#define           JOINTS_NUM_GROUPS     4

typedef enum {JOINTS_GRP0=0,JOINTS_GRP1=1,JOINTS_GRP2=2,JOINTS_GRP3=2} joints_grp_t;

class CDarwinRobot
{
  private:
    CDynamixelServerFTDI *dyn_server;  
    CDynamixel *robot_device;
    std::string robot_name;
    unsigned char robot_id;
    // motion manager variables
    unsigned char manager_num_servos;
    unsigned char manager_status;
    /* action status */
    unsigned char action_status;
    /* walk status */
    unsigned char walk_status;
    /* joint group status */
    unsigned char joints_status[JOINTS_NUM_GROUPS];
    /* head tracking status */
    unsigned char head_tracking_status;
  public:
    CDarwinRobot(const std::string &name,std::string &bus_id,int bus_speed, unsigned char id);
    // GPIO interface
    void gpio_set_led(led_t led);
    void gpio_clear_led(led_t led);
    void gpio_toggle_led(led_t led);
    void gpio_blink_led(led_t led, int period_ms);
    void gpio_aux1_color(double red, double green, double blue);
    void gpio_aux2_color(double red, double green, double blue);
    bool is_button_pressed(pushbutton_t pushbutton);
    // ADC interface
    void adc_start(void);
    bool adc_is_running(void);
    void adc_set_period(unsigned char period_ms);
    double adc_get_value(adc_t adc);
    double adc_get_temperature(void);
    double adc_get_vrefint(void);
    void adc_stop(void);
    // imu interface
    void imu_start(void);
    void imu_stop(void);
    void imu_start_gyro_cal(void);
    void imu_set_cal_samples(unsigned short int num_samples);
    unsigned short int imu_get_cal_samples(void);
    bool imu_is_gyro_cal_done(void);
    bool imu_is_accel_detected(void);
    bool imu_is_gyro_detected(void);
    void imu_get_accel_data(double *x,double *y,double *z);
    void imu_get_gyro_data(double *x,double *y,double *z);
    // motion manager interface
    void mm_set_period(double period_ms);
    double mm_get_period(void);
    unsigned char mm_get_num_servos(void);
    unsigned int mm_get_present_servos(void);
    void mm_start(void);
    void mm_stop(void);
    bool mm_is_running(void);
    void mm_enable_balance(void);
    void mm_disable_balance(void);
    bool mm_is_balance_enabled(void);
    void mm_enable_power(void);
    void mm_disable_power(void);
    bool mm_is_power_enabled(void);
    void mm_enable_servo(unsigned char servo_id);
    void mm_disable_servo(unsigned char servo_id);
    bool mm_is_servo_enabled(unsigned char servo_id);
    void mm_assign_module(unsigned char servo_id, mm_mode_t mode);
    void mm_assign_module(std::string &servo,std::string &module);
    mm_mode_t mm_get_module(unsigned char servo_id);
    std::vector<double> mm_get_servo_angles(void);
    double mm_get_servo_angle(unsigned char servo_id);
    void mm_set_servo_offset(unsigned char servo_id,double offset);
    double mm_get_servo_offset(unsigned char servo_id);
    std::vector<double> mm_get_servo_offsets(void);
    void mm_load_config(std::string &filename);
    void mm_set_balance_gains(double knee,double ankle_pitch,double hip_roll,double ankle_roll);
    void mm_get_balance_gains(double *knee,double *ankle_pitch,double *hip_roll,double *ankle_roll);
    // motion action interface
    void action_load_page(unsigned char page_id);
    unsigned char action_get_current_page(void);
    unsigned char action_get_current_step(void);
    void action_start(void);
    void action_stop(void);
    bool action_is_page_running(void);
    /* walking interface */
    void walk_set_x_offset(double offset_m);
    double walk_get_x_offset(void);
    void walk_set_y_offset(double offset_m);
    double walk_get_y_offset(void);
    void walk_set_z_offset(double offset_m);
    double walk_get_z_offset(void);
    void walk_set_roll_offset(double offset_rad);
    double walk_get_roll_offset(void);
    void walk_set_pitch_offset(double offset_rad);
    double walk_get_pitch_offset(void);
    void walk_set_yaw_offset(double offset_rad);
    double walk_get_yaw_offset(void);
    void walk_set_hip_pitch_offset(double offset_rad);
    double walk_get_hip_pitch_offset(void);
    void walk_set_period(double period_ms);
    double walk_get_period(void);
    void walk_set_ssp_dsp_ratio(double ratio);
    double walk_get_ssp_dsp_ratio(void);
    void walk_set_fwd_bwd_ratio(double ratio);
    double walk_get_fwd_bwd_ratio(void);
    void walk_set_foot_height(double height_m);
    double walk_get_foot_height(void);
    void walk_set_left_right_swing(double swing_m);
    double walk_get_left_right_swing(void);
    void walk_set_top_down_swing(double swing_m);
    double walk_get_top_down_swing(void);
    void walk_set_pelvis_offset(double offset_rad);
    double walk_get_pelvis_offset(void);
    void walk_set_arm_swing_gain(double gain);
    double walk_get_arm_swing_gain(void);
    void walk_set_trans_speed(double speed_m_s);
    double walk_get_trans_speed(void);
    void walk_set_rot_speed(double speed_rad_s);
    double walk_get_rot_speed(void);
    void walk_start(void);
    void walk_stop(void);
    bool is_walking(void);
    void walk_set_x_step(double step_m);
    double walk_get_x_step(void);
    void walk_set_y_step(double step_m);
    double walk_get_y_step(void);
    void walk_set_turn_step(double step_rad);
    double walk_get_turn_step(void);
    // joint motion interface
    void joints_load(joints_grp_t group,std::vector<unsigned char> &servos,std::vector<double> &angles,std::vector<double> &speeds,std::vector<double> &accels);
    std::vector<unsigned char> joints_get_group_servos(joints_grp_t group);
    std::vector<double> joints_get_group_angles(joints_grp_t group);
    std::vector<double> joints_get_group_speeds(joints_grp_t group);
    std::vector<double> joints_get_group_accels(joints_grp_t group);
    void joints_start(joints_grp_t group);
    void joints_stop(joints_grp_t group);
    bool joints_are_moving(joints_grp_t group);
    // head tracking interface
    void head_set_pan_pid(double p,double i,double d,double i_clamp);
    void head_get_pan_pid(double *p,double *i,double *d,double *i_clamp);
    void head_set_tilt_pid(double p,double i,double d,double i_clamp);
    void head_get_tilt_pid(double *p,double *i,double *d,double *i_clamp);
    void head_set_pan_range(double max,double min);
    void head_get_pan_range(double *max,double *min);
    void head_set_tilt_range(double max,double min);
    void head_get_tilt_range(double *max,double *min);
    void head_start_tracking(double pan,double tilt);
    void head_stop_tracking(void);
    bool head_is_tracking(void);
    void head_set_new_target(double pan,double tilt);
    void head_get_current_target(double *pan,double *tilt);
    ~CDarwinRobot();
};

#endif
